/**
  Copyright (C) 1998 XYZ, Inc. All rights Reserved.
  XXX   %Z%%M% version %I% date:%G%
  Component:  SPKIPolicy
  Package: policy
  File:    SPKIPolicy.java
*/

package fi.hut.tcm.policy;
import fi.hut.tcm.spki.impl.TagSimple;
import fi.hut.tcm.spki.impl.TagStar;
import fi.hut.tcm.spki.impl.TagExpr;
import fi.hut.tcm.spki.impl.TagString;
import fi.hut.tcm.spki.impl.Issuer;
import fi.hut.tcm.spki.impl.Subject;
import fi.hut.tcm.spki.Composite;
import fi.hut.tcm.spki.Tag;
import fi.hut.tcm.spki.ByteString;
import fi.hut.tcm.spki.SPKICertificate;
import fi.hut.tcm.repository.*;
import java.security.*;
import java.util.*;
import java.net.URL;
import sun.security.util.PropertyExpander;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;

/**
  SPKIPolicy
  
  The SPKIPolicy uses SPKICertificates to express and manage the security 
  policy.
*/
public class SPKIPolicy
  extends Policy {

  private boolean debug = true;

  /**
    The Self key is the public key of the Verifier,  i.e. this Policy 
    object.
  */
  private PublicKey self;
  
  /**
    The CertificateRepository object used by this Policy to retrieve 
    certificates.
  */
  private SPKICertificateRepository storage;
  
  /**
    The default permissions are equal to the SandBox.
  */
  private Permissions defaultPerm;
  
  /**
    Indicates whether the policy object has been initialized or not.
  */
  private boolean initialized = false;
  
  /**
    private static final Class[] PARAMS = { String.class, String.class};
  */
  private static Class[] PARAMS = {String.class, String.class};
  
  /**
   */
  private Issuer selfIssuer;

  /**
    Creates a SPKIPolicy object.
  */
  public SPKIPolicy() {
  
  //+ SPKIPolicy::SPKIPolicy( )
    super();

    initialize();
  //- SPKIPolicy::SPKIPolicy( )
  }
  
  /**
    Initializes the policy object by fetching the Self key into memory, 
    finding the certificate repository and creating the default policy.
  */
  private void initialize() {
  
  //+ SPKIPolicy::initialize( )

    // Initializing the Self key.
    // FIX: KeyStore file name???
    // Should we do some checks?

    String ksfile;

    try {
      ksfile = (String) AccessController.doPrivileged(new PrivilegedAction() {
	public Object run() {
	  return Security.getProperty("selfkey.file");
	};
      });
      if (ksfile == null) {
	ksfile = PropertyExpander.expand("file:${user.home}/.keystore");
	// ksfile = PropertyExpander.expand("file:${java.home}/lib/security/SelfKey.dat");
      }
      URL keyStoreUrl = new URL(ksfile);
      InputStream streamIn = keyStoreUrl.openStream(); 
      KeyStore ks = KeyStore.getInstance("JKS");
      // FIX: password should not be hard coded!
      ks.load(streamIn, "passwd".toCharArray());
      streamIn.close();
      java.security.cert.Certificate c = ks.getCertificate("Self");
      self = c.getPublicKey();
    } catch (Exception e) {
      // This is BAD. What should we do? 
    }

    defaultPerm = new Permissions();
    // Create the default Permissions for any ProtectionDomain
    defaultPerm.add(new PropertyPermission("java.version","read"));
    defaultPerm.add(new PropertyPermission("java.vendor","read"));
    defaultPerm.add(new PropertyPermission("java.vendor.url","read"));
    defaultPerm.add(new PropertyPermission("java.class.version","read"));
    defaultPerm.add(new PropertyPermission("os.name","read"));
    defaultPerm.add(new PropertyPermission("os.version","read"));
    defaultPerm.add(new PropertyPermission("os.arch","read"));
    defaultPerm.add(new PropertyPermission("file.separator","read"));
    defaultPerm.add(new PropertyPermission("path.separator","read"));
    defaultPerm.add(new PropertyPermission("line.separator","read"));

    // Create the self issuer
    selfIssuer = new Issuer(self);

    // Initialize the certStorage
    try {
      storage = SPKICertificateRepository.getInstance();
      storage.load();
    } catch (Exception e) {
      System.err.println("Policy unable to initialize CertificateRepository: " + e.toString());
    }

    initialized = true;
  //- SPKIPolicy::initialize( )
  }
  
  /**
    Refreshes the given policy object.
  */
  public void refresh() {
  
  //+ SPKIPolicy::refresh( )
    initialize();
  //- SPKIPolicy::refresh( )
  }
  
  /**
    Evaluates the policy object with the specified CodeSource, and creates 
    a Permissions object with the set of permissions for that principal's 
    protection domain.
  */
  public PermissionCollection getPermissions(CodeSource codesource) {
  
  //+ SPKIPolicy::getPermissions(CodeSource)
    DynamicPermissions perm = new DynamicPermissions(codesource);
    return perm;
  //- SPKIPolicy::getPermissions(CodeSource)
  }
  
  /**
  */
  public boolean checkPermission(Permission perm, CodeSource cs) {

  //+ SPKIPolicy::checkPermission(Permission, CodeSource)
    java.security.cert.Certificate certs[] = cs.getCertificates();

    // Java extensions should get all permissions regardless 
    // of the keys that signed them (if any).
    // FIX: or should they?
    try {
      String extdir = 
	PropertyExpander.expand("file:${java.home}/lib/ext/");
      URL extensions = new URL(extdir);
      if (cs.getLocation().equals(extensions)) {
	if (debug) 
	  System.out.println("Granted: class is java extension");
	return true;
      }
    } catch (Exception e) {
      // FIX: This is bad. Do something drastic!
    }
    
    if (!initialized) {
      initialize();
    }
    
    // If default permissions are enough return true
    if (defaultPerm.implies(perm)) {
      if (debug) 
	System.out.println("Granted: default permission");
      return true;
    }
    // If they are no Certificates
    if (certs.length == 0) {
      if (debug) 
	System.out.println("Denied: no certificates");
      return false;
    }
    
    // Only take SPKICertificates into account
    SPKICertificate temp[] = new SPKICertificate[certs.length];
    int j = 0;
    for (int i = 0; i < certs.length; i++) {
      if (certs[i] instanceof SPKICertificate) {
	temp[j] = (SPKICertificate) certs[i];
	j++;
      }
    }
    if ((j > 0) && (j < certs.length)) {
      certs = new SPKICertificate[j];
      System.arraycopy(temp, 0, certs, 0, certs.length);
    } else if (j <= 0) {
      // If they are no SPKICertificates
      if (debug) 
	System.out.println("Denied: no spki certificates");
      return false;
    } 
    
    // Evaluate the policy.
    // FIX: this cannot handle whole certificate chains that are 
    //      delivered with the class!
    int i = 0;
    SPKICertificate c;
    while (i < certs.length) {
      c = (SPKICertificate) certs[i];
      //      resolvePerms(c.getPublicKey(), p);
      if (resolvePerms(c, perm)) {
	System.out.println("Grant ok");
	return true;
      }
      i++;
    }

    System.out.println("Denied: fall through");
    return false;
  //- SPKIPolicy::checkPermission(Permission, CodeSource)
  }

  /**
  */
  private boolean resolvePerms(SPKICertificate spkic, Permission perm) {
  //+ SPKIPolicy::resolvePerms(SPKICertificate, Permission)

    SPKICertificate c;
    boolean result = false;
    Vector v;
    int i = 0;
    
    try {

      // Check if certificate is valid. Don't check delegation or 
      // signature, since this may be the last certificate in a chain,
      // which is already verified and does not need delegation.
      if (spkic.isValid()) {
	
	Permission test = tagToPermission(spkic.getTag());
	if (debug) 
	  System.out.println("Asked permission: " + perm.toString() + 
			     ", perm to test: " + test.toString());
	if (!test.implies(perm)) {
	  if (debug) 
	    System.out.println("Denied: no imply");
	  return false;
	}
	
	if (spkic.getIssuer().equals(selfIssuer.getPublicKey())) {
	  if (debug) 
	    System.out.println("Granted: issuer is self");
	  return true;
	}
	
	try {
	  if (storage == null) {
	    if (debug)
	      System.out.println("Denied: CertificateRepository was not initialized.");
	    return false;
	  }
	  v = storage.getCertsBySubject(new Subject(spkic.getIssuer()));
	} catch (Exception e) {
	  // Unable to find certificates: warn the user and return.
	  System.err.println("Policy cannot find certificates: " + e.toString());
	  return false;
	}
	if (v == null) {
	  if (debug)
	    System.out.println("Denied: There are no more certificates to resolve");
	  return false;  // There are no certificates issued to the key
	}

	while (i < v.size()) {
	  c = (SPKICertificate)v.get(i);
	  c.verify(c.getPublicKey());
	  if (c.getDeleg()) {
	    if (resolvePerms(c, perm)) {
	      System.out.println("Grant ok");
	      return true; 
	    }
	  }
	  i++;
	}
      } else
	if (debug) 
	  System.out.println("Denied: not valid");
      
      return false;
    } catch (Exception e) {
      e.printStackTrace();
      return false;
    }

  //+ SPKIPolicy::resolvePerms(SPKICertificate, Permission)
  }

  /**
    Interprets the Tag into a Java Permission object. 
  */
  private Permission tagToPermission(Tag tag) {

  //+ SPKIPolicy::tagToPermission(Tag)
    TagStar star;
    TagSimple simple;
    String type = null;
    String target = null;
    String action = null;
    
    if (debug) 
      System.out.println("Tag: " + tag.toString());
    try {
      star = (TagStar) tag.get(new TagStar());
      if (star != null)
	return new AllPermission();
    } catch (Exception e) {
    }
    
    try {
      TagExpr tmp = (TagExpr) tag.get(0);
      simple = (TagSimple) tmp.get(0);
      ByteString bs = (ByteString) simple.get(0);
      if (!bs.getValueString().equals("java-permission")) {
	if (debug) 
	  System.out.println("Tag is not a java-permission tag");
	return null;
      }
      TagSimple typeTag = null;
      TagSimple targetTag = null;
      TagSimple actionTag = null;
      int i = simple.size();
      typeTag = (TagSimple) ((Composite) simple.get(1)).get(0);
      TagString str = (TagString) ((Composite) typeTag.get(1)).get(0);
      type = ((ByteString) str.get(0)).getValueString();
      if (debug) 
	System.out.println("foo");
      if (i >= 3) {
	targetTag = (TagSimple) ((Composite) simple.get(2)).get(0);
	str = (TagString) ((Composite) targetTag.get(1)).get(0);
	target = ((ByteString) str.get(0)).getValueString();
      } 
      if (i == 4) {
	actionTag = (TagSimple) ((Composite) simple.get(3)).get(0);
	str = (TagString) ((Composite) actionTag.get(1)).get(0);
	action = ((ByteString) str.get(0)).getValueString();
      }
      if (i > 4) {
	if (debug) 
	  System.out.println("Tag has too many fields, ignoring");
	return null;
      }
      return getPermInstance(type, target, action);
    } catch (Exception e) {
      System.err.println("Exception in trying to make TagSimple: " + e.toString());
    }
    
    // FIX: handle other types of tags
    return null;
	
  //- SPKIPolicy::tagToPermission(Tag)
  }

  /**
    final
  */
  private static final Permission getPermInstance(String type, String target, String action)
       throws ClassNotFoundException, InstantiationException, 
	 IllegalAccessException, NoSuchMethodException, InvocationTargetException  {
  //+ SPKIPolicy::getPermInstance(String, String, String)

	   Class xxx = Class.forName(type);
	   java.lang.reflect.Constructor c = xxx.getConstructor(PARAMS);
	   return (Permission) c.newInstance(new Object[] { target, action });
	   
  //- SPKIPolicy::getPermInstance(String, String, String)
  }
  
  /*
    Finds all permissions that have been given to the key  through a valid 
    certificate chain, and adds them to the Permissions object.
  */
    /*
  private void resolvePerms(Key key, Permissions perm) {
  
  //+ SPKIPolicy::resolvePerms(Key, Permissions)
    Vector v;
    Permission p = null;
    SPKICertificate c;
    PublicKey issuer = null;
    Tag t;
    int i = 0;

  //- SPKIPolicy::resolvePerms(Key, Permissions)
  }
  */
  /*
    A recursive method to go through the certificate graph. Only keys are 
    allowed as certificate issuers. <br><b>Note:</b><br> At the moment, 
    loops (i.e. A delegates the right X to B,  who delegates it back to A) 
    in the graph cause the algorithm to enter an infinite loop.
  */
    /*
  private void reduceCertChain(Key key, Permission testp, Permissions perm) {
  
  //+ SPKIPolicy::reduceCertChain(Key, Permission, Permissions)
  //- SPKIPolicy::reduceCertChain(Key, Permission, Permissions)
  }
  */
} // SPKIPolicy

// +*************** UNUSED CODE ***************
// -*************** UNUSED CODE ***************
