/**
  Copyright (C) 1998 XYZ, Inc. All rights Reserved.
  XXX   %Z%%M% version %I% date:%G%
  Component:  SpkiParserVisitor
  Package: impl
  File:    SpkiParserVisitor.java
*/

package fi.hut.tcm.spki.impl;
import fi.hut.tcm.spki.*;
import fi.hut.tcm.spki.format.*;
import java.io.ByteArrayInputStream;
import java.util.*;
import java.math.BigInteger;

/**
  SpkiParserVisitor
  
  Creates and traverses Generic Format based on parsing SPKI Canonical 
  Format.
*/
public class SpkiParserVisitor
  extends Object
  implements ComponentVisitor {

  /**
  */
  private static final int LOOK_AHEAD = 10;
  
  /**
  */
  private ByteArrayInputStream bin;
  
  /**
  */
  public SpkiParserVisitor() {
  
  //+ SpkiParserVisitor::SpkiParserVisitor( )
    super();
  //- SpkiParserVisitor::SpkiParserVisitor( )
  }
  
  /**
  */
  public GenericFormat run(CanonicalSpki convertThis)
    throws IllegalArgumentException  {
  
  //+ SpkiParserVisitor::run(CanonicalSpki)
      GenericFormat g;

      bin = new ByteArrayInputStream(convertThis.getBuffer());
      String[] tmp = { "sequence", "cert" };
      try {
	g = (GenericFormat) recurse(alternation(tmp));
      } catch (NullPointerException e) {
	// Assumed to have come from alternation( ) returning null in an
	// unexpected point resulting at recurse(null)
	throw new IllegalArgumentException();
      }
      return g;
  //- SpkiParserVisitor::run(CanonicalSpki)
  }
  
  /**
  */
  public void att(Sequence node) {
  
  //+ SpkiParserVisitor::att(Sequence)
      Composite c;
      String next;
      String[] tmp = { "cert", "public-key", "signature", "do" };

      while ((c = (Composite) alternation(tmp)) != null) {
	if (c instanceof GeneralOp)
	  if ((next = readString()).equals("hash"))
	    c = new HashOp();
	  else
	    c.add(new ByteString(next));
	node.add(recurse(c));
      }
  //- SpkiParserVisitor::att(Sequence)
  }
  
  /**
  */
  public void att(SPart node) {
  
  //+ SpkiParserVisitor::att(SPart)
      Component s;
      
      // Java API is inconsistent, but mark(int) really should mean this
      bin.mark(LOOK_AHEAD);
      if ((char) bin.read() == '(')
	s = new SExpr();
      else
	s = new ByteString();
      bin.reset();
      node.add(recurse(s));
  //- SpkiParserVisitor::att(SPart)
  }
  
  /**
  */
  public void att(Hash node) {
  
  //+ SpkiParserVisitor::att(Hash)
      node.add(recurse(new HashAlgName()));
      node.add(recurse(new HashValue()));

      bin.mark(LOOK_AHEAD);
      if ((char) bin.read() != ')') {
	bin.reset();
	node.add(recurse(new Uri()));
	skipClosingParentheses();
      }
  //- SpkiParserVisitor::att(Hash)
  }
  
  /**
  */
  public void att(SExpr node) {
  
  //+ SpkiParserVisitor::att(SExpr)
      skipOpeningParentheses();
      node.add(recurse(new ByteString()));

      bin.mark(LOOK_AHEAD);
      while ((char) bin.read() != ')') {
	bin.reset();
	node.add(recurse(new SPart()));
	bin.mark(LOOK_AHEAD);
      }
  //- SpkiParserVisitor::att(SExpr)
  }
  
  /**
  */
  public void att(HashAlgName node) {
  
  //+ SpkiParserVisitor::att(HashAlgName)
      bin.mark(LOOK_AHEAD); 
      if ((char) bin.read() == '[') {
	bin.reset();
	node.add(recurse(new Uri()));
      }
      else {
	bin.reset();
	String s = readString();
	if (s.equals("md5") ||
	    s.equals("sha1"))
	  node.add(new ByteString(s));
	else {
	  try {
	    Uri u = new Uri();
	    u.setValue(s);
	    node.add(u);
	  } catch (SPKIException e) {
	    // XXX 
	    System.err.println("Exception: " + e);
	  }
	}
      }
  //- SpkiParserVisitor::att(HashAlgName)
  }
  
  /**
  */
  public void att(Issuer node) {
  
  //+ SpkiParserVisitor::att(Issuer)
      String[] tmp = { "name", "public-key", "hash" };
      Component child = alternation(tmp);
      if (child instanceof Hash)
	child = new HashOfKey();
      node.add(recurse(child));
      skipClosingParentheses();
      // XXX name should contain only one namestep
  //- SpkiParserVisitor::att(Issuer)
  }
  
  /**
  */
  public void att(Subject node) {
  
  //+ SpkiParserVisitor::att(Subject)
      String[] tmp = { "name", "public-key", "hash", "object-hash", 
		       "secret-key", "keyholder", "k-of-n" };
      node.add(recurse(alternation(tmp)));
      skipClosingParentheses();
  //- SpkiParserVisitor::att(Subject)
  }
  
  /**
  */
  public void att(GeneralOp node) {
  
  //+ SpkiParserVisitor::att(GeneralOp)
      // The ByteString is already read
      bin.mark(LOOK_AHEAD);
      while ((char) bin.read() != ')') {
	bin.reset();
	node.add(recurse(new SPart()));
	bin.mark(LOOK_AHEAD);
      }
  //- SpkiParserVisitor::att(GeneralOp)
  }
  
  /**
  */
  public void att(HashOp node) {
  
  //+ SpkiParserVisitor::att(HashOp)
      node.add(recurse(new HashAlgName()));
      skipClosingParentheses();
  //- SpkiParserVisitor::att(HashOp)
  }
  
  /**
  */
  public void att(Name node) {
  
  //+ SpkiParserVisitor::att(Name)
      bin.mark(LOOK_AHEAD);
      if ((char) bin.read() == '(') {
	bin.reset();
	String[] tmp = { "public-key", "hash" };
	Component child = alternation(tmp);
	if (child instanceof Hash)
	  child = new HashOfKey();
	node.add(recurse(child));
      }
      else
	bin.reset();

      bin.mark(LOOK_AHEAD);
      while ((char) bin.read() != ')') {
	bin.reset();
	node.add(recurse(new ByteString()));
	bin.mark(LOOK_AHEAD);
      }
  //- SpkiParserVisitor::att(Name)
  }
  
  /**
  */
  public void att(Keyholder node) {
  
  //+ SpkiParserVisitor::att(Keyholder)
      String[] tmp = { "public-key", "hash" };
      Component child = alternation(tmp);
      if (child instanceof Hash)
	child = new HashOfKey();
      node.add(recurse(child));
      skipClosingParentheses();
  //- SpkiParserVisitor::att(Keyholder)
  }
  
  /**
  */
  public void att(ThresholdSubj node) {
  
  //+ SpkiParserVisitor::att(ThresholdSubj)
      String[] tmp = { "subject" };

      node.add(recurse(new KVal()));
      int nVal = new BigInteger(readString().getBytes()).intValue();
      for (int i = 0; i < nVal; i++)
	node.add(recurse(alternation(tmp)));
      skipClosingParentheses();
  //- SpkiParserVisitor::att(ThresholdSubj)
  }
  
  /**
  */
  public void att(SecKey node) {
  
  //+ SpkiParserVisitor::att(SecKey)
      char next;

      node.add(recurse(new SecSigAlgId()));

      bin.mark(LOOK_AHEAD);
      while ((next = (char) bin.read()) != ')')
	if (next == '(') {
	  bin.reset();
	  node.add(recurse(new SExpr()));
	  bin.mark(LOOK_AHEAD);
	}
	else {
	  bin.reset();
	  node.add(recurse(new Uri()));
	  skipClosingParentheses();
	  break;
	}
  //- SpkiParserVisitor::att(SecKey)
  }
  
  /**
  */
  public void att(PubKey node) {
  
  //+ SpkiParserVisitor::att(PubKey)
      char next;
      
      node.add(recurse(new PubSigAlgId()));
      
      bin.mark(LOOK_AHEAD);
      while ((next = (char) bin.read()) != ')')
	if (next == '(') {
	  bin.reset();
	  node.add(recurse(new SExpr()));
	  bin.mark(LOOK_AHEAD);
	}
	else {
	  bin.reset();
	  node.add(recurse(new Uri()));
	  skipClosingParentheses();
	  break;
	}
  //- SpkiParserVisitor::att(PubKey)
  }
  
  /**
  */
  public void att(PubSigAlgId node) {
  
  //+ SpkiParserVisitor::att(PubSigAlgId)
      bin.mark(LOOK_AHEAD); 
      if ((char) bin.read() == '[') {
	bin.reset();
	node.add(recurse(new Uri()));
      }
      else {
	bin.reset();
	String s = readString();
	if (s.equals("rsa-pkcs1-md5") ||
	    s.equals("rsa-pkcs1-sha1") ||
	    s.equals("dsa-sha1"))
	  node.add(new ByteString(s));
	else {
	  try {
	    Uri u = new Uri();
	    u.setValue(s);
	    node.add(u);
	  } catch (SPKIException e) {
	    // XXX 
	    System.err.println("Exception: " + e);
	  }
	}
      }
  //- SpkiParserVisitor::att(PubSigAlgId)
  }
  
  /**
  */
  public void att(SecSigAlgId node) {
  
  //+ SpkiParserVisitor::att(SecSigAlgId)
      bin.mark(LOOK_AHEAD); 
      if ((char) bin.read() == '[') {
	bin.reset();
	node.add(recurse(new Uri()));
      }
      else {
	bin.reset();
	String s = readString();
	if (s.equals("hmac-md5") ||
	    s.equals("hmac-sha1") ||
	    s.equals("des-cbc-mac"))
	  node.add(new ByteString(s));
	else {
	  try {
	    Uri u = new Uri();
	    u.setValue(s);
	    node.add(u);
	  } catch (SPKIException e) {
	    // XXX 
	    System.err.println("Exception: " + e);
	  }
	}
      }
  //- SpkiParserVisitor::att(SecSigAlgId)
  }
  
  /**
  */
  public void att(OnlineTest node) {
  
  //+ SpkiParserVisitor::att(OnlineTest)
      node.add(recurse(new OnlineType()));
      node.add(recurse(new Uri()));

      String[] tmp = { "public-key", "hash" };
      Component child = alternation(tmp);
      if (child instanceof Hash)
	child = new HashOfKey();
      node.add(recurse(child));

      bin.mark(LOOK_AHEAD);
      while ((char) bin.read() != ')') {
	bin.reset();
	node.add(recurse(new SPart()));
	bin.mark(LOOK_AHEAD);
      }
  //- SpkiParserVisitor::att(OnlineTest)
  }
  
  /**
  */
  public void att(ObjHash node) {
  
  //+ SpkiParserVisitor::att(ObjHash)
      String[] tmp = { "hash" };
      node.add(recurse(alternation(tmp)));
      skipClosingParentheses();
  //- SpkiParserVisitor::att(ObjHash)
  }
  
  /**
  */
  public void att(Tag node) {
  
  //+ SpkiParserVisitor::att(Tag)
      bin.mark(LOOK_AHEAD);
      if ((char) bin.read() == '(')
	if ((char) bin.read() == '1')
	  if ((char) bin.read() == ':')
	    if ((char) bin.read() == '*')
	      if ((char) bin.read() == ')') {
		node.add(new TagStar());
		skipClosingParentheses();
		return;
	      }

      bin.reset();
      node.add(recurse(new TagExpr()));
      skipClosingParentheses();
  //- SpkiParserVisitor::att(Tag)
  }
  
  /**
  */
  public void att(Sig node) {
  
  //+ SpkiParserVisitor::att(Sig)
      String[] tmp = { "hash" };
      String[] tmp2 = { "public-key", "hash" };

      node.add(recurse(alternation(tmp)));

      Component child = alternation(tmp2);
      if (child instanceof Hash)
	child = new HashOfKey();
      node.add(recurse(child));

      node.add(recurse(new SigVal()));
      skipClosingParentheses();
  //- SpkiParserVisitor::att(Sig)
  }
  
  /**
  */
  public void att(Cert node) {
  
  //+ SpkiParserVisitor::att(Cert)
      Vector temp = new Vector();
      Component c;
      boolean deleg = false;

      String[] fields = { "version", "display", "issuer", "issuer-info", 
			  "subject", "subject-info", "propagate", "tag",
			  "not-before", "not-after", "online", "comment" };
      while ((c = alternation(fields)) != null) {
	if (c instanceof SkipSExpr)
	  moveToEndOfSExpr();
	else
	  temp.addElement(recurse(c));
	if (c instanceof Deleg)
	  deleg = true;
      }
      if (!deleg) {
	try {
	  Deleg d = new Deleg();
	  d.setValue("f");
	  temp.addElement(d);
	} catch (SPKIException e) {
	  // XXX 
	  System.err.println("Exception: " + e);
	}
      }
      // Reorder components and move not-before, not-after & online under Valid
      Vector ordered = new Vector();
      for (int i = 0; i < fields.length; i++)
	for (int j = 0; j < temp.size(); j++)
	  if (getInstance(fields[i]).getClass().isInstance(temp.elementAt(j)))
	    if (fields[i].equals("not-before") ||
		fields[i].equals("not-after") ||
		fields[i].equals("online")) {
	      if (!(ordered.lastElement() instanceof Valid))
		ordered.addElement(new Valid());
	      ((Composite) ordered.lastElement()).add
		((Component) temp.elementAt(j));
	    }
	    else
	      ordered.addElement(temp.elementAt(j));
      temp = ordered;
      // Move all under node
      for (Enumeration e = temp.elements(); e.hasMoreElements(); )
	node.add((Component) e.nextElement());
      // XXX Is it ok to drop locations completely?
  //- SpkiParserVisitor::att(Cert)
  }
  
  /**
  */
  public void att(Valid node) {
  
  //+ SpkiParserVisitor::att(Valid)
      // Valid is done in att(Cert)
      throw new Error("We should not get to att(Valid) in SPKI Parsing.");
  //- SpkiParserVisitor::att(Valid)
  }
  
  /**
  */
  public void att(NotBefore node) {
  
  //+ SpkiParserVisitor::att(NotBefore)
      String[] tmp = { "hash" };

      node.add(recurse(alternation(tmp)));
      skipClosingParentheses();
  //- SpkiParserVisitor::att(NotBefore)
  }
  
  /**
  */
  public void att(NotAfter node) {
  
  //+ SpkiParserVisitor::att(NotAfter)
      String[] tmp = { "hash" };

      node.add(recurse(alternation(tmp)));
      skipClosingParentheses();
  //- SpkiParserVisitor::att(NotAfter)
  }
  
  /**
  */
  public void att(Deleg node) {
  
  //+ SpkiParserVisitor::att(Deleg)
    try {
      node.setValue("t");
    } catch (SPKIException e) {
      // XXX 
      System.err.println("Exception: " + e);
    }
      skipClosingParentheses();
  //- SpkiParserVisitor::att(Deleg)
  }
  
  /**
  */
  public void att(Comment node) {
  
  //+ SpkiParserVisitor::att(Comment)
      att((ByteString) node);
      skipClosingParentheses();
  //- SpkiParserVisitor::att(Comment)
  }
  
  /**
  */
  public void att(KVal node) {
  
  //+ SpkiParserVisitor::att(KVal)
      att((ByteString) node);
  //- SpkiParserVisitor::att(KVal)
  }
  
  /**
  */
  public void att(Uri node) {
  
  //+ SpkiParserVisitor::att(Uri)
      att((ByteString) node);
  //- SpkiParserVisitor::att(Uri)
  }
  
  /**
  */
  public void att(OnlineType node)
    throws IllegalArgumentException  {
  
  //+ SpkiParserVisitor::att(OnlineType)
	String s = readString();
	if (s.equals("crl") ||
	    s.equals("reval") ||
	    s.equals("one-time")) {
	  try {
	    node.setValue(s);
	  } catch (SPKIException e) {
	    throw new IllegalArgumentException(e.toString());
	  }
	}
	else
	  throw new IllegalArgumentException();
  //- SpkiParserVisitor::att(OnlineType)
  }
  
  /**
  */
  public void att(HashValue node) {
  
  //+ SpkiParserVisitor::att(HashValue)
      att((ByteString) node);
  //- SpkiParserVisitor::att(HashValue)
  }
  
  /**
  */
  public void att(Time node) {
  
  //+ SpkiParserVisitor::att(Time)
      att((ByteString) node);
  //- SpkiParserVisitor::att(Time)
  }
  
  /**
  */
  public void att(HashOfKey node) {
  
  //+ SpkiParserVisitor::att(HashOfKey)
      String[] tmp = { "hash" };

      alternation(tmp);
      att((Hash) node);
  //- SpkiParserVisitor::att(HashOfKey)
  }
  
  /**
  */
  public void att(SigVal node) {
  
  //+ SpkiParserVisitor::att(SigVal)
      att((SPart) node);
  //- SpkiParserVisitor::att(SigVal)
  }
  
  /**
  */
  public void att(ByteString node)
    throws IllegalArgumentException  {
  
  //+ SpkiParserVisitor::att(ByteString)
      bin.mark(LOOK_AHEAD);
      if ((char) bin.read() == '[') {
	node.setDisplayHint(readString());
	if ((char) bin.read() != ']')
	  throw new IllegalArgumentException();
      }
      else
	bin.reset();
      try {
	node.setValue(readString());
      } catch (SPKIException e) {
	throw new IllegalArgumentException(e.toString());
      }
  //- SpkiParserVisitor::att(ByteString)
  }
  
  /**
  */
  public void att(Lte node)
    throws IllegalArgumentException  {
  
  //+ SpkiParserVisitor::att(Lte)
      String s = readString();
      if (s.equals("l") ||
	  s.equals("le")) {
	try {
	  node.setValue(s);
	} catch (SPKIException e) {
	  throw new IllegalArgumentException(e.toString());
	}
      } else
	throw new IllegalArgumentException("Str:" + s + " len:" + s.length());
  //- SpkiParserVisitor::att(Lte)
  }
  
  /**
  */
  public void att(Gte node)
    throws IllegalArgumentException  {
  
  //+ SpkiParserVisitor::att(Gte)
      String s = readString();
      if (s.equals("g") ||
	  s.equals("ge")) {
	try {
	  node.setValue(s);
	} catch (SPKIException e) {
	  throw new IllegalArgumentException(e.toString());
	}
      } else
	throw new IllegalArgumentException("Str:" + s + " len:" + s.length());
  //- SpkiParserVisitor::att(Gte)
  }
  
  /**
  */
  public void att(RangeOrdering node)
    throws IllegalArgumentException  {
  
  //+ SpkiParserVisitor::att(RangeOrdering)
      String s = readString();
      if (s.equals("alpha") ||
	  s.equals("numeric") ||
	  s.equals("time") ||
	  s.equals("binary") ||
	  s.equals("date")) {
	try {
	  node.setValue(s);
	} catch (SPKIException e) {
	  throw new IllegalArgumentException(e.toString());
	}
      } else
	throw new IllegalArgumentException("Str:" + s);
  //- SpkiParserVisitor::att(RangeOrdering)
  }
  
  /**
  */
  public void att(TagStar node) {
  
  //+ SpkiParserVisitor::att(TagStar)
      // Handled at Tag
      throw new Error(); // We should never get here
  //- SpkiParserVisitor::att(TagStar)
  }
  
  /**
  */
  public void att(UpLim node) {
  
  //+ SpkiParserVisitor::att(UpLim)
      node.add(recurse(new Lte()));
      node.add(recurse(new ByteString()));
  //- SpkiParserVisitor::att(UpLim)
  }
  
  /**
  */
  public void att(LowLim node) {
  
  //+ SpkiParserVisitor::att(LowLim)
      node.add(recurse(new Gte()));
      node.add(recurse(new ByteString()));
  //- SpkiParserVisitor::att(LowLim)
  }
  
  /**
  */
  public void att(TagRange node) {
  
  //+ SpkiParserVisitor::att(TagRange)
      node.add(recurse(new RangeOrdering()));
      for (int i = 0; i < 2; i++) {
	bin.mark(LOOK_AHEAD);
	char nextChar = (char) bin.read();
	bin.reset();
	if (nextChar != ')') {
	  bin.mark(LOOK_AHEAD);
	  String comparator = readString();
	  bin.reset();
	  if (comparator.equals("ge") || comparator.equals("g")) {
	    if (node.size() > 1) { // UpLim already read
	      Component temp = node.get(1); // Reorder them
	      node.remove(temp);
	      node.add(recurse(new LowLim()));
	      node.add(temp);
	    }
	    else
	      node.add(recurse(new LowLim()));
	  }
	  else
	    node.add(recurse(new UpLim()));
	}
      }
      skipClosingParentheses();
  //- SpkiParserVisitor::att(TagRange)
  }
  
  /**
  */
  public void att(TagPrefix node) {
  
  //+ SpkiParserVisitor::att(TagPrefix)
      node.add(recurse(new ByteString()));
      skipClosingParentheses();
  //- SpkiParserVisitor::att(TagPrefix)
  }
  
  /**
  */
  public void att(TagSet node) {
  
  //+ SpkiParserVisitor::att(TagSet)
      bin.mark(LOOK_AHEAD);
      while ((char) bin.read() != ')') {
	bin.reset();
	node.add(recurse(new TagExpr()));
	bin.mark(LOOK_AHEAD);
      }
  //- SpkiParserVisitor::att(TagSet)
  }
  
  /**
  */
  public void att(TagString node) {
  
  //+ SpkiParserVisitor::att(TagString)
      // This was handled at att(TagExpr)
      throw new Error(); // We should never get here
  //- SpkiParserVisitor::att(TagString)
  }
  
  /**
  */
  public void att(TagSimple node) {
  
  //+ SpkiParserVisitor::att(TagSimple)
      node.add(recurse(new ByteString()));
      bin.mark(LOOK_AHEAD);
      while ((char) bin.read() != ')') {
	bin.reset();
	node.add(recurse(new TagExpr()));
	bin.mark(LOOK_AHEAD);
      }
  //- SpkiParserVisitor::att(TagSimple)
  }
  
  /**
  */
  public void att(TagExpr node)
    throws IllegalArgumentException  {
  
  //+ SpkiParserVisitor::att(TagExpr)
      TagString tag;

      bin.mark(LOOK_AHEAD);
      try {
	skipOpeningParentheses();
      } catch (IllegalArgumentException e) { // We will skip att(TagString)
	bin.reset();
	tag = new TagString();
	tag.add(recurse(new ByteString()));
	node.add(tag);
	return;
      }
      
      bin.mark(LOOK_AHEAD);
      if ((char) bin.read() == '1')
	if ((char) bin.read() == ':')
	  if ((char) bin.read() == '*') {
	    String str = readString();
	    if (str.equals("range")) { // We will skip att(TagString)
	      tag = new TagString();
	      tag.add(recurse(new TagRange()));
	      node.add(tag);
	      return;
	    }
	    if (str.equals("prefix")) { // We will skip att(TagString)
	      tag = new TagString();
	      tag.add(recurse(new TagPrefix()));
	      node.add(tag);
	      return;
	    }
	    if (str.equals("set")) {
	      node.add(recurse(new TagSet()));
	      return;
	    }
	    throw new IllegalArgumentException("Str:" + str);
	  }
      // TagSimple
      bin.reset();
      node.add(recurse(new TagSimple()));
  //- SpkiParserVisitor::att(TagExpr)
  }
  
  /**
  */
  public void at(Object node) {

  //+ ComponentVisitor::at(Object)
    if (node.getClass().getName().equals("fi.hut.tcm.spki.ByteString"))
      att((ByteString) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Cert"))
      att((Cert) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Comment"))
      att((Comment) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Deleg"))
      att((Deleg) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.GeneralOp"))
      att((GeneralOp) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Gte"))
      att((Gte) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Hash"))
      att((Hash) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.HashAlgName"))
      att((HashAlgName) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.HashOfKey"))
      att((HashOfKey) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.HashOp"))
      att((HashOp) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.HashValue"))
      att((HashValue) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Issuer"))
      att((Issuer) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Keyholder"))
      att((Keyholder) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.KVal"))
      att((KVal) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.LowLim"))
      att((LowLim) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Lte"))
      att((Lte) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Name"))
      att((Name) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.NotAfter"))
      att((NotAfter) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.NotBefore"))
      att((NotBefore) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.ObjHash"))
      att((ObjHash) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.OnlineTest"))
      att((OnlineTest) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.OnlineType"))
      att((OnlineType) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.PubKey"))
      att((PubKey) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.PubSigAlgId"))
      att((PubSigAlgId) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.RangeOrdering"))
      att((RangeOrdering) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.SecKey"))
      att((SecKey) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.SecSigAlgId"))
      att((SecSigAlgId) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.format.Sequence"))
      att((Sequence) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.SExpr"))
      att((SExpr) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Sig"))
      att((Sig) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.SigVal"))
      att((SigVal) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.SPart"))
      att((SPart) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Subject"))
      att((Subject) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Tag"))
      att((Tag) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.TagExpr"))
      att((TagExpr) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.TagPrefix"))
      att((TagPrefix) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.TagRange"))
      att((TagRange) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.TagSet"))
      att((TagSet) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.TagSimple"))
      att((TagSimple) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.TagStar"))
      att((TagStar) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.TagString"))
      att((TagString) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.ThresholdSubj"))
      att((ThresholdSubj) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Time"))
      att((Time) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.UpLim"))
      att((UpLim) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Uri"))
      att((Uri) node);
    if (node.getClass().getName().equals("fi.hut.tcm.spki.impl.Valid"))
      att((Valid) node);
  //- ComponentVisitor::at(Object)
  }
  
  /**
    Tests if CanonicalSpki has one of the choices as the next s-expression.
     If it does, returns a component of the corresponding type. Returns 
    null if the list was over and throws IllegalArgumentException if match 
    was not found.
  */
  private Component alternation(String[] choices)
    throws IllegalArgumentException  {
  
  //+ SpkiParserVisitor::alternation(String[])
      String next;
      Component c;

      if ((next = readType()) == null)
	return null;
      
      for (int i = 0; i < choices.length; i++)
	if (next.equals(choices[i])) {
	  c = getInstance(next);
	  if (c == null)
	    throw new Error(); // We really should get a match.
	  else
	    return c;
	}
      throw new IllegalArgumentException();
  //- SpkiParserVisitor::alternation(String[])
  }
  
  /**
    Gets an instance of a class that matches the given string indentifier. 
    Returns null if match was not found.
  */
  private Component getInstance(String type) {
  
  //+ SpkiParserVisitor::getInstance(String)
      try {
	return (Component)
	  ((Class) CanonicalSpki.parseassociations.get(type)).newInstance();
      } catch (IllegalAccessException e) {
	throw new InstantiationError();
      } catch (InstantiationException e) {
	throw new InstantiationError();
      }
  //- SpkiParserVisitor::getInstance(String)
  }
  
  /**
    Reads the type of the following s-expression. Returns null if the first
     byte was ')', meaning the list was over. Throws an exception for 
    unexpected bytes in the input.
  */
  private String readType()
    throws IllegalArgumentException  {
  
  //+ SpkiParserVisitor::readType( )
      char next;
      
      if ((next = (char) bin.read()) != '(')
	if (next == ')')
	  return null;
	else
	  throw new IllegalArgumentException
	    ("Instead of a parentheses, the first byte was: " + next);

      return readString();
  //- SpkiParserVisitor::readType( )
  }
  
  /**
    Returns following string. Throws an exception for unexpected input.
  */
  private String readString()
    throws IllegalArgumentException  {
  
  //+ SpkiParserVisitor::readString( )
      int len = 0;
      char next;

      while ((next = (char) bin.read()) != ':')
	if (Character.isDigit(next))
	  len = len * 10 + Character.digit(next, 10);
	else
	  throw new IllegalArgumentException
	    ("NotDigit:" + next + " (int):" + (int) next + " Len:" + len);

      byte[] result = new byte[len];
      bin.read(result, 0, len);
      return new String(result);
  //- SpkiParserVisitor::readString( )
  }
  
  /**
  */
  private Component recurse(Component into) {
  
  //+ SpkiParserVisitor::recurse(Component)
      into.accept((Visitor) this);
      return into;
  //- SpkiParserVisitor::recurse(Component)
  }
  
  /**
  */
  private void moveToEndOfSExpr() {
  
  //+ SpkiParserVisitor::moveToEndOfSExpr( )
      char next;

      while ((next = (char) bin.read()) != ')')
	if (next == '(')
	  moveToEndOfSExpr();
  //- SpkiParserVisitor::moveToEndOfSExpr( )
  }
  
  /**
  */
  private void skipClosingParentheses()
    throws IllegalArgumentException  {
  
  //+ SpkiParserVisitor::skipClosingParentheses( )
      int next = bin.read();

      if ((char) next != ')')
	throw new IllegalArgumentException
	  ("NotClosing:" + (char) next + " (int):" + next);
  //- SpkiParserVisitor::skipClosingParentheses( )
  }
  
  /**
  */
  private void skipOpeningParentheses()
    throws IllegalArgumentException  {
  
  //+ SpkiParserVisitor::skipOpeningParentheses( )
      int next = bin.read();

      if ((char) next != '(')
	throw new IllegalArgumentException
	  ("NotOpening" + (char) next + " (int):" + next);
  //- SpkiParserVisitor::skipOpeningParentheses( )
  }
  
} // SpkiParserVisitor

// +*************** UNUSED CODE ***************
// -*************** UNUSED CODE ***************
